@@ -28,8 +28,9 @@ from collections import deque
 from textwrap import dedent
 
 from logbook.base import (
-    CRITICAL, ERROR, WARNING, NOTICE, INFO, DEBUG, NOTSET, level_name_property,
-    _missing, lookup_level, Flags, ContextObject, ContextStackManager)
+    CRITICAL, ERROR, WARNING, NOTICE, INFO, DEBUG, TRACE, NOTSET, level_name_property,
+    _missing, lookup_level, Flags, ContextObject, ContextStackManager,
+    _datetime_factory)
 from logbook.helpers import (
     rename, b, _is_text_stream, is_unicode, PY2, zip, xrange, string_types,
     integer_types, reraise, u, with_metaclass)
@@ -497,8 +498,7 @@ class LimitingHandlerMixin(HashingHandlerMixin):
 
             if (not suppression_count and
                     len(self._record_limits) >= self.max_record_cache):
-                cache_items = self._record_limits.items()
-                cache_items.sort()
+                cache_items = sorted(self._record_limits.items())
                 del cache_items[:int(self._record_limits)
                                 * self.record_cache_prune]
                 self._record_limits = dict(cache_items)
@@ -650,11 +650,9 @@ class FileHandler(StreamHandler):
 
 class GZIPCompressionHandler(FileHandler):
     def __init__(self, filename, encoding=None, level=NOTSET,
-                 format_string=None, delay=False, filter=None, bubble=False,
-                 compression_window_size=4*1024**2, compression_quality=9):
+                 format_string=None, delay=False, filter=None, bubble=False, compression_quality=9):
+
         self._compression_quality = compression_quality
-        self._compression_window_size = compression_window_size
-        self._num_unflushed_bytes = 0
         super(GZIPCompressionHandler, self).__init__(filename, mode='wb', encoding=encoding, level=level,
                              format_string=format_string, delay=delay, filter=filter, bubble=bubble)
 
@@ -668,14 +666,11 @@ class GZIPCompressionHandler(FileHandler):
             item = item.encode(encoding=self.encoding)
         self.ensure_stream_is_open()
         self.stream.write(item)
-        self._num_unflushed_bytes += len(item)
 
     def should_flush(self):
-        return self._num_unflushed_bytes >= self._compression_window_size
-
-    def flush(self):
-        super(GZIPCompressionHandler, self).flush()
-        self._num_unflushed_bytes = 0
+        # gzip manages writes independently. Flushing prematurely could mean
+        # duplicate flushes and thus bloated files
+        return False
 
 
 class BrotliCompressionHandler(FileHandler):
@@ -867,29 +862,72 @@ class TimedRotatingFileHandler(FileHandler):
 
     By default it will keep all these files around, if you want to limit
     them, you can specify a `backup_count`.
+
+    You may supply an optional `rollover_format`. This allows you to specify
+    the format for the filenames of rolled-over files.
+    the format as
+
+    So for example if you configure your handler like this::
+
+        handler = TimedRotatingFileHandler(
+            '/var/log/foo.log',
+            date_format='%Y-%m-%d',
+            rollover_format='{basename}{ext}.{timestamp}')
+
+    The filenames for the logfiles will look like this::
+
+        /var/log/foo.log.2010-01-10
+        /var/log/foo.log.2010-01-11
+        ...
+
+    Finally, an optional argument `timed_filename_for_current` may be set to
+    false if you wish to have the current log file match the supplied filename
+    until it is rolled over
     """
 
     def __init__(self, filename, mode='a', encoding='utf-8', level=NOTSET,
                  format_string=None, date_format='%Y-%m-%d',
-                 backup_count=0, filter=None, bubble=False):
-        FileHandler.__init__(self, filename, mode, encoding, level,
-                             format_string, True, filter, bubble)
+                 backup_count=0, filter=None, bubble=False,
+                 timed_filename_for_current=True,
+                 rollover_format='{basename}-{timestamp}{ext}'):
         self.date_format = date_format
         self.backup_count = backup_count
-        self._fn_parts = os.path.splitext(os.path.abspath(filename))
-        self._filename = None
 
-    def _get_timed_filename(self, datetime):
-        return (datetime.strftime('-' + self.date_format)
-                .join(self._fn_parts))
+        self.rollover_format = rollover_format
 
-    def should_rollover(self, record):
-        fn = self._get_timed_filename(record.time)
-        rv = self._filename is not None and self._filename != fn
-        # remember the current filename.  In case rv is True, the rollover
-        # performing function will already have the new filename
-        self._filename = fn
-        return rv
+        self.original_filename = filename
+        self.basename, self.ext = os.path.splitext(os.path.abspath(filename))
+        self.timed_filename_for_current = timed_filename_for_current
+
+        self._timestamp = self._get_timestamp(_datetime_factory())
+        if self.timed_filename_for_current:
+            filename = self.generate_timed_filename(self._timestamp)
+        elif os.path.exists(filename):
+            self._timestamp = self._get_timestamp(
+                datetime.fromtimestamp(
+                    os.stat(filename).st_mtime
+                )
+            )
+
+        FileHandler.__init__(self, filename, mode, encoding, level,
+                             format_string, True, filter, bubble)
+
+    def _get_timestamp(self, datetime):
+        """
+        Fetches a formatted string witha timestamp of the given datetime
+        """
+        return datetime.strftime(self.date_format)
+
+    def generate_timed_filename(self, timestamp):
+        """
+        Produces a filename that includes a timestamp in the format supplied
+        to the handler at init time.
+        """
+        timed_filename = self.rollover_format.format(
+            basename=self.basename,
+            timestamp=timestamp,
+            ext=self.ext)
+        return timed_filename
 
     def files_to_delete(self):
         """Returns a list with the files that have to be deleted when
@@ -897,10 +935,14 @@ class TimedRotatingFileHandler(FileHandler):
         """
         directory = os.path.dirname(self._filename)
         files = []
+        rollover_regex = re.compile(self.rollover_format.format(
+            basename=re.escape(self.basename),
+            timestamp='.+',
+            ext=re.escape(self.ext),
+        ))
         for filename in os.listdir(directory):
             filename = os.path.join(directory, filename)
-            if (filename.startswith(self._fn_parts[0] + '-') and
-                    filename.endswith(self._fn_parts[1])):
+            if rollover_regex.match(filename):
                 files.append((os.path.getmtime(filename), filename))
         files.sort()
         if self.backup_count > 1:
@@ -908,19 +950,34 @@ class TimedRotatingFileHandler(FileHandler):
         else:
             return files[:]
 
-    def perform_rollover(self):
-        self.stream.close()
+    def perform_rollover(self, new_timestamp):
+        if self.stream is not None:
+            self.stream.close()
+
+        if (
+                not self.timed_filename_for_current
+                and os.path.exists(self._filename)
+        ):
+            filename = self.generate_timed_filename(self._timestamp)
+            os.rename(self._filename, filename)
+
         if self.backup_count > 0:
             for time, filename in self.files_to_delete():
                 os.remove(filename)
+
+        if self.timed_filename_for_current:
+            self._filename = self.generate_timed_filename(new_timestamp)
+        self._timestamp = new_timestamp
+
         self._open('w')
 
     def emit(self, record):
         msg = self.format(record)
         self.lock.acquire()
         try:
-            if self.should_rollover(record):
-                self.perform_rollover()
+            new_timestamp = self._get_timestamp(record.time)
+            if new_timestamp != self._timestamp:
+                self.perform_rollover(new_timestamp)
             self.write(self.encode(msg))
             self.flush()
         finally:
@@ -1005,6 +1062,11 @@ class TestHandler(Handler, StringFormatterHandlerMixin):
         """`True` if any :data:`DEBUG` records were found."""
         return any(r.level == DEBUG for r in self.records)
 
+    @property
+    def has_traces(self):
+        """`True` if any :data:`TRACE` records were found."""
+        return any(r.level == TRACE for r in self.records)
+
     def has_critical(self, *args, **kwargs):
         """`True` if a specific :data:`CRITICAL` log record exists.
 
@@ -1053,6 +1115,14 @@ class TestHandler(Handler, StringFormatterHandlerMixin):
         kwargs['level'] = DEBUG
         return self._test_for(*args, **kwargs)
 
+    def has_trace(self, *args, **kwargs):
+        """`True` if a specific :data:`TRACE` log record exists.
+
+        See :ref:`probe-log-records` for more information.
+        """
+        kwargs['level'] = TRACE
+        return self._test_for(*args, **kwargs)
+
     def _test_for(self, message=None, channel=None, level=None):
         def _match(needle, haystack):
             """Matches both compiled regular expressions and strings"""
@@ -1464,7 +1534,7 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
     def __init__(self, application_name=None, address=None,
                  facility='user', socktype=socket.SOCK_DGRAM,
                  level=NOTSET, format_string=None, filter=None,
-                 bubble=False):
+                 bubble=False, record_delimiter=None):
         Handler.__init__(self, level, filter, bubble)
         StringFormatterHandlerMixin.__init__(self, format_string)
         self.application_name = application_name
@@ -1475,14 +1545,24 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
             else:
                 address = '/dev/log'
 
-        self.address = address
+        self.remote_address = self.address = address
         self.facility = facility
         self.socktype = socktype
 
         if isinstance(address, string_types):
             self._connect_unixsocket()
+            self.enveloper = self.unix_envelope
+            default_delimiter = u'\x00'
         else:
             self._connect_netsocket()
+            self.enveloper = self.net_envelope
+            default_delimiter = u'\n'
+
+        self.record_delimiter = default_delimiter \
+            if record_delimiter is None else record_delimiter
+
+        self.connection_exception = getattr(
+            __builtins__, 'BrokenPipeError', socket.error)
 
     def _connect_unixsocket(self):
         self.unixsocket = True
@@ -1498,7 +1578,7 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
         self.unixsocket = False
         self.socket = socket.socket(socket.AF_INET, self.socktype)
         if self.socktype == socket.SOCK_STREAM:
-            self.socket.connect(self.address)
+            self.socket.connect(self.remote_address)
             self.address = self.socket.getsockname()
 
     def encode_priority(self, record):
@@ -1507,15 +1587,44 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
                                                self.LOG_WARNING)
         return (facility << 3) | priority
 
-    def emit(self, record):
-        prefix = u('')
-        if self.application_name is not None:
-            prefix = self.application_name + u(':')
-        self.send_to_socket((u('<%d>%s%s\x00') % (
+    def wrap_segments(self, record, before):
+        msg = self.format(record)
+        segments = [segment for segment in msg.split(self.record_delimiter)]
+        return (before + segment + self.record_delimiter
+                for segment in segments)
+        
+    def unix_envelope(self, record):
+        before = u'<{}>{}'.format(
             self.encode_priority(record),
-            prefix,
-            self.format(record)
-        )).encode('utf-8'))
+            self.application_name + ':' if self.application_name else '')
+        return self.wrap_segments(record, before)
+
+    def net_envelope(self, record):
+        # Gross but effective
+        try:
+            format_string = self.format_string
+            application_name = self.application_name
+            if not application_name and record.channel and \
+               '{record.channel}: ' in format_string:
+                self.format_string = format_string.replace(
+                    '{record.channel}: ', '')
+                self.application_name = record.channel
+            # RFC 5424: <PRIVAL>version timestamp hostname app-name procid
+            #           msgid structured-data message
+            before = u'<{}>1 {}Z {} {} {} - - '.format(
+                self.encode_priority(record),
+                record.time.isoformat(),
+                socket.gethostname(),
+                self.application_name if self.application_name else '-',
+                record.process)
+            return self.wrap_segments(record, before)
+        finally:
+            self.format_string = format_string
+            self.application_name = application_name
+
+    def emit(self, record):
+        for segment in self.enveloper(record):
+            self.send_to_socket(segment.encode('utf-8'))
 
     def send_to_socket(self, data):
         if self.unixsocket:
@@ -1528,7 +1637,11 @@ class SyslogHandler(Handler, StringFormatterHandlerMixin):
             # the flags are no longer optional on Python 3
             self.socket.sendto(data, 0, self.address)
         else:
-            self.socket.sendall(data)
+            try:
+                self.socket.sendall(data)
+            except self.connection_exception:
+                self._connect_netsocket()
+                self.socket.send(data)
 
     def close(self):
         self.socket.close()
@@ -1787,6 +1900,10 @@ class GroupHandler(WrapperHandler):
         Handler.pop_thread(self)
         self.rollover()
 
+    def pop_context(self):
+        Handler.pop_context(self)
+        self.rollover()
+
     def pop_greenlet(self):
         Handler.pop_greenlet(self)
         self.rollover()
